from django.contrib.contenttypes.fields import GenericRel, GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models

from jet_django.filters.model import model_filter_class_factory
from jet_django.serializers.model import model_serializer_factory
from jet_django.serializers.model_detail import model_detail_serializer_factory
from jet_django.views.model import model_viewset_factory


class JetAdminModelDescription(object):
    def __init__(self, Model, fields=None, hidden=False):
        self.model = Model
        self.fields = fields
        self.hidden = hidden

        self.content_type = ContentType.objects.get_for_model(Model)
        self.field_names = list(map(lambda x: x.name, self.get_display_model_fields()))
        self.serializer = model_serializer_factory(Model, self.field_names)
        self.detail_serializer = model_detail_serializer_factory(Model, self.field_names)
        self.filter_class = model_filter_class_factory(Model, self.get_display_model_fields(), self.get_model_relations())
        self.queryset = Model.objects.all()
        self.viewset = model_viewset_factory(
            Model,
            self.filter_class,
            self.serializer,
            self.detail_serializer,
            self.queryset
        )

    @property
    def viewset_url(self):
        return 'models/(?P<model>{};{})'.format(self.content_type.app_label, self.content_type.model)

    def get_model_fields(self):
        fields = self.model._meta.get_fields()
        def filter_fields(x):
            if any(map(lambda rel: isinstance(x, rel), [
                models.OneToOneRel,
                models.OneToOneField,
                models.ManyToOneRel,
                models.ManyToManyField,
                models.ManyToManyRel,
                GenericRel,
                GenericForeignKey,
                GenericRelation
            ])):
                return False
            return True
        return filter(filter_fields, fields)

    def get_model_relations(self):
        fields = self.model._meta.get_fields(include_hidden=True)
        def filter_fields(x):
            if any(map(lambda rel: isinstance(x, rel), [
                models.OneToOneRel,
                models.OneToOneField,
                models.ManyToOneRel,
                models.ManyToManyField,
                models.ManyToManyRel
            ])):
                return True
            return False
        return list(filter(filter_fields, fields))

    def get_model_relation_through(self, field):
        if isinstance(field, models.ManyToManyRel):
            return self.serialize_model(field.through)
        elif isinstance(field, models.ManyToManyField):
            return self.serialize_model(field.remote_field.through)

    def get_display_model_fields(self):
        fields = self.get_model_fields()
        def filter_fields(x):
            if self.fields:
                return x.name in self.fields
            return True
        return filter(filter_fields, fields)

    def serialize(self):
        return {
            'model': '{};{}'.format(self.content_type.app_label, self.content_type.model),
            'db_table': self.model._meta.db_table,
            'verbose_name': self.model._meta.verbose_name,
            'verbose_name_plural': self.model._meta.verbose_name_plural,
            'hidden': self.hidden,
            'fields': map(lambda field: {
                'name': field.name,
                'db_column': field.get_attname_column()[1],
                'verbose_name': field.verbose_name,
                'is_relation': field.is_relation,
                'field': field.__class__.__name__,
                'required': not field.blank,
                'null': field.null,
                'editable': field.editable,
                'filterable': field.name in self.filter_class.Meta.fields,
                'params': {
                    'related_model': self.serialize_model(field.related_model)
                }
            }, self.get_display_model_fields()),
            'flex_fields': [],
            'relations': map(lambda field: {
                'name': field.name,
                'verbose_name': field.related_model._meta.verbose_name_plural,
                'related_model': self.serialize_model(field.related_model),
                'field': field.__class__.__name__,
                'related_model_field': field.remote_field.name,
                'through': self.get_model_relation_through(field)
            }, self.get_model_relations())
        }

    def get_model(self):
        return {
            'model': '{};{}'.format(self.content_type.app_label, self.content_type.model)
        }

    def get_related_models(self):
        return map(lambda field: {
            'model': field.related_model,
            'model_info': self.serialize_model(field.related_model)
        }, self.get_model_relations())

    def serialize_model(self, Model):
        if not Model:
            return
        content_type = ContentType.objects.get_for_model(Model)
        return {
            'model': '{};{}'.format(content_type.app_label, content_type.model),
        }



from django.db import connection, DatabaseError
from django.utils import six
from django.utils.translation import ugettext_lazy as _

from jet_django.deps.rest_framework import serializers, status
from jet_django.deps.rest_framework.exceptions import ValidationError, APIException


class ParamsSerializers(serializers.CharField):
    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        return list(filter(lambda x: x != '', value.split(',')))

    def to_representation(self, value):
        return list(value)


class SqlError(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _('Query failed')
    default_code = 'invalid'

    def __init__(self, detail):
        self.detail = {'error': str(detail)}

    def __str__(self):
        return six.text_type(self.detail)


class SqlSerializer(serializers.Serializer):
    query = serializers.CharField()
    params = ParamsSerializers(required=False)

    def validate_query(self, value):
        forbidden = ['insert', 'update', 'delete', 'grant', 'show']
        for i in range(len(forbidden)):
            forbidden.append('({}'.format(forbidden[i]))
        if any(map(lambda x: ' {} '.format(value.lower()).find(' {} '.format(x)) != -1, forbidden)):
            raise ValidationError('forbidden query')
        return value

    def execute(self, data):
        with connection.cursor() as cursor:
            try:
                cursor.execute(data['query'], data.get('params', []))
            except (DatabaseError, TypeError) as e:
                raise SqlError(e)

            rows = cursor.fetchall()

            def map_column(x):
                if not isinstance(x, tuple) or len(x) == 0:
                    return
                if x[0] == '?column?':
                    return
                return x[0]

            return {'data': rows, 'columns': map(map_column, cursor.description)}


class SqlsSerializer(serializers.Serializer):
    queries = SqlSerializer(many=True)

    def execute(self, data):
        serializer = SqlSerializer()
        return map(lambda x: serializer.execute(x), data['queries'])



from django.core.management import BaseCommand

from jet_django.utils.backend import get_token


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('token', nargs='?', type=str)

    def handle(self, *args, **options):
        token = get_token()

        if token:
            print('Jet Admin Token:')
            print(token)
        else:
            print('Jet Admin Token is not set')



from jet_django.deps.rest_framework import serializers

from jet_django.models.widget import Widget


class WidgetDetailSerializer(serializers.ModelSerializer):
    class Meta:
        model = Widget
        fields = (
            'id',
            'widget_type',
            'name',
            'x',
            'y',
            'width',
            'height',
            'params',
            'date_add',
            'dashboard'
        )



import logging
import requests

from jet_django import settings
from jet_django.models.token import Token

logger = logging.getLogger('jet_django')


def api_method_url(method):
    return '{}/{}'.format(settings.JET_BACKEND_API_BASE_URL, method)


def get_token():
    token = Token.objects.all().first()
    return token.token if token else None


def register_token():
    token = Token.objects.all().first()

    if token:
        logger.info('[JET] Token already registered')
        return token, False

    url = api_method_url('project_tokens/')
    headers = {
        'User-Agent': 'Jet Django'
    }

    r = requests.request('POST', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        logger.error('[JET] Register Token request error: %s %s %s', r.status_code, r.reason, r.text)
        return None, False

    result = r.json()
    token = Token.objects.create(token=result['token'], date_add=result['date_add'])

    return token, True


def is_token_activated(token):
    url = api_method_url('project_tokens/{}/'.format(token.token))
    headers = {
        'User-Agent': 'Jet Django'
    }

    r = requests.request('GET', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        return False

    result = r.json()

    return result.get('activated') is True


def reset_token():
    Token.objects.all().delete()

    return register_token()


def project_auth(token, permission=None):
    project_token = Token.objects.all().first()

    if not project_token:
        logger.error('[JET] Project Auth request error: not token registered')
        return {
            'result': False
        }

    url = api_method_url('project_auth/')
    data = {
        'project_token': project_token.token,
        'token': token
    }
    headers = {
        'User-Agent': 'Jet Django'
    }

    if permission:
        data.update(permission)

    r = requests.request('POST', url, data=data, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        logger.error('[JET] Project Auth request error: %s %s %s', r.status_code, r.reason, r.text)
        return {
            'result': False
        }

    result = r.json()

    if result.get('access_disabled'):
        logger.error('[JET] Project Auth request error: access_disabled')
        return {
            'result': False,
            'warning': result.get('warning')
        }

    return {
        'result': True,
        'warning': result.get('warning')
    }



import json
from django.utils.translation import ugettext_lazy as _

from jet_django.deps.rest_framework import serializers


class RawFileField(serializers.CharField):

    def __init__(self, *args, **kwargs):
        if 'validators' in kwargs:
            del kwargs['validators']

        self.default_error_messages['invalid_format'] = _('Not a valid format.')
        super().__init__(*args, **kwargs)

    def to_internal_value(self, data):
        if isinstance(data, dict):
            obj = data
        else:
            try:
                obj = json.loads(data)
            except (ValueError, TypeError):
                self.fail('invalid_format')

        try:
            return obj['value']
        except KeyError:
            self.fail('invalid_format')

    def to_representation(self, value):
        return {
            'value': super().to_representation(value),
            'url': serializers.FileField.to_representation(self, value)
        }



from collections import namedtuple
from datetime import datetime, time

from django import forms
from django.utils.dateparse import parse_datetime
from django.utils.encoding import force_str
from django.utils.translation import ugettext_lazy as _

from .conf import settings
from .constants import EMPTY_VALUES
from .utils import handle_timezone
from .widgets import (
    BaseCSVWidget,
    CSVWidget,
    DateRangeWidget,
    LookupChoiceWidget,
    RangeWidget
)


class RangeField(forms.MultiValueField):
    widget = RangeWidget

    def __init__(self, fields=None, *args, **kwargs):
        if fields is None:
            fields = (
                forms.DecimalField(),
                forms.DecimalField())
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        if data_list:
            return slice(*data_list)
        return None


class DateRangeField(RangeField):
    widget = DateRangeWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.DateField(),
            forms.DateField())
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        if data_list:
            start_date, stop_date = data_list
            if start_date:
                start_date = handle_timezone(
                    datetime.combine(start_date, time.min),
                    False
                )
            if stop_date:
                stop_date = handle_timezone(
                    datetime.combine(stop_date, time.max),
                    False
                )
            return slice(start_date, stop_date)
        return None


class DateTimeRangeField(RangeField):
    widget = DateRangeWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.DateTimeField(),
            forms.DateTimeField())
        super().__init__(fields, *args, **kwargs)


class TimeRangeField(RangeField):
    widget = DateRangeWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.TimeField(),
            forms.TimeField())
        super().__init__(fields, *args, **kwargs)


class Lookup(namedtuple('Lookup', ('value', 'lookup_expr'))):
    def __new__(cls, value, lookup_expr):
        if value in EMPTY_VALUES or lookup_expr in EMPTY_VALUES:
            raise ValueError(
                "Empty values ([], (), {}, '', None) are not "
                "valid Lookup arguments. Return None instead."
            )

        return super().__new__(cls, value, lookup_expr)


class LookupChoiceField(forms.MultiValueField):
    default_error_messages = {
        'lookup_required': _('Select a lookup.'),
    }

    def __init__(self, field, lookup_choices, *args, **kwargs):
        empty_label = kwargs.pop('empty_label', settings.EMPTY_CHOICE_LABEL)
        fields = (field, ChoiceField(choices=lookup_choices, empty_label=empty_label))
        widget = LookupChoiceWidget(widgets=[f.widget for f in fields])
        kwargs['widget'] = widget
        kwargs['help_text'] = field.help_text
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        if len(data_list) == 2:
            value, lookup_expr = data_list
            if value not in EMPTY_VALUES:
                if lookup_expr not in EMPTY_VALUES:
                    return Lookup(value=value, lookup_expr=lookup_expr)
                else:
                    raise forms.ValidationError(
                        self.error_messages['lookup_required'],
                        code='lookup_required')
        return None


class IsoDateTimeField(forms.DateTimeField):
    """
    Supports 'iso-8601' date format too which is out the scope of
    the ``datetime.strptime`` standard library

    # ISO 8601: ``http://www.w3.org/TR/NOTE-datetime``

    Based on Gist example by David Medina https://gist.github.com/copitux/5773821
    """
    ISO_8601 = 'iso-8601'
    input_formats = [ISO_8601]

    def strptime(self, value, format):
        value = force_str(value)

        if format == self.ISO_8601:
            parsed = parse_datetime(value)
            if parsed is None:  # Continue with other formats if doesn't match
                raise ValueError
            return handle_timezone(parsed)
        return super().strptime(value, format)


class BaseCSVField(forms.Field):
    """
    Base field for validating CSV types. Value validation is performed by
    secondary base classes.

    ex::
        class IntegerCSVField(BaseCSVField, filters.IntegerField):
            pass

    """
    base_widget_class = BaseCSVWidget

    def __init__(self, *args, **kwargs):
        widget = kwargs.get('widget') or self.widget
        kwargs['widget'] = self._get_widget_class(widget)

        super().__init__(*args, **kwargs)

    def _get_widget_class(self, widget):
        # passthrough, allows for override
        if isinstance(widget, BaseCSVWidget) or (
                isinstance(widget, type) and
                issubclass(widget, BaseCSVWidget)):
            return widget

        # complain since we are unable to reconstruct widget instances
        assert isinstance(widget, type), \
            "'%s.widget' must be a widget class, not %s." \
            % (self.__class__.__name__, repr(widget))

        bases = (self.base_widget_class, widget, )
        return type(str('CSV%s' % widget.__name__), bases, {})

    def clean(self, value):
        if value is None:
            return None
        return [super(BaseCSVField, self).clean(v) for v in value]


class BaseRangeField(BaseCSVField):
    # Force use of text input, as range must always have two inputs. A date
    # input would only allow a user to input one value and would always fail.
    widget = CSVWidget

    default_error_messages = {
        'invalid_values': _('Range query expects two values.')
    }

    def clean(self, value):
        value = super().clean(value)

        assert value is None or isinstance(value, list)

        if value and len(value) != 2:
            raise forms.ValidationError(
                self.error_messages['invalid_values'],
                code='invalid_values')

        return value


class ChoiceIterator(object):
    # Emulates the behavior of ModelChoiceIterator, but instead wraps
    # the field's _choices iterable.

    def __init__(self, field, choices):
        self.field = field
        self.choices = choices

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)
        if self.field.null_label is not None:
            yield (self.field.null_value, self.field.null_label)

        # Python 2 lacks 'yield from'
        for choice in self.choices:
            yield choice

    def __len__(self):
        add = 1 if self.field.empty_label is not None else 0
        add += 1 if self.field.null_label is not None else 0
        return len(self.choices) + add


class ModelChoiceIterator(forms.models.ModelChoiceIterator):
    # Extends the base ModelChoiceIterator to add in 'null' choice handling.
    # This is a bit verbose since we have to insert the null choice after the
    # empty choice, but before the remainder of the choices.

    def __iter__(self):
        iterable = super().__iter__()

        if self.field.empty_label is not None:
            yield next(iterable)
        if self.field.null_label is not None:
            yield (self.field.null_value, self.field.null_label)

        # Python 2 lacks 'yield from'
        for value in iterable:
            yield value

    def __len__(self):
        add = 1 if self.field.null_label is not None else 0
        return super().__len__() + add


class ChoiceIteratorMixin(object):
    def __init__(self, *args, **kwargs):
        self.null_label = kwargs.pop('null_label', settings.NULL_CHOICE_LABEL)
        self.null_value = kwargs.pop('null_value', settings.NULL_CHOICE_VALUE)

        super().__init__(*args, **kwargs)

    def _get_choices(self):
        return super()._get_choices()

    def _set_choices(self, value):
        super()._set_choices(value)
        value = self.iterator(self, self._choices)

        self._choices = self.widget.choices = value
    choices = property(_get_choices, _set_choices)


# Unlike their Model* counterparts, forms.ChoiceField and forms.MultipleChoiceField do not set empty_label
class ChoiceField(ChoiceIteratorMixin, forms.ChoiceField):
    iterator = ChoiceIterator

    def __init__(self, *args, **kwargs):
        self.empty_label = kwargs.pop('empty_label', settings.EMPTY_CHOICE_LABEL)
        super().__init__(*args, **kwargs)


class MultipleChoiceField(ChoiceIteratorMixin, forms.MultipleChoiceField):
    iterator = ChoiceIterator

    def __init__(self, *args, **kwargs):
        self.empty_label = None
        super().__init__(*args, **kwargs)


class ModelChoiceField(ChoiceIteratorMixin, forms.ModelChoiceField):
    iterator = ModelChoiceIterator

    def to_python(self, value):
        # bypass the queryset value check
        if self.null_label is not None and value == self.null_value:
            return value
        return super().to_python(value)


class ModelMultipleChoiceField(ChoiceIteratorMixin, forms.ModelMultipleChoiceField):
    iterator = ModelChoiceIterator

    def _check_values(self, value):
        null = self.null_label is not None and value and self.null_value in value
        if null:  # remove the null value and any potential duplicates
            value = [v for v in value if v != self.null_value]

        result = list(super()._check_values(value))
        result += [self.null_value] if null else []
        return result



from jet_django.deps.rest_framework import views
from jet_django.deps.rest_framework.response import Response
from jet_django.mixins.cors_api_view import CORSAPIViewMixin
from jet_django.permissions import HasProjectPermissions
from jet_django.serializers.file_upload import FileUploadSerializer


class FileUploadView(CORSAPIViewMixin, views.APIView):
    authentication_classes = ()
    permission_classes = (HasProjectPermissions,)

    def post(self, request, *args, **kwargs):
        serializer = FileUploadSerializer(data=request.data, context={'request': self.request})
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response(serializer.data)
